/*
 * Stefan Ioana, 336CB
 * PM, 2018
 * Line Follower
 */

#include <avr/io.h>
#include <avr/interrupt.h>

#include <util/delay.h>

#include <stdlib.h>
#include <stdio.h>

/* Reference tention for ADC. */
#define ADC_AREF_VOLTAGE 5

/* Reference difference to compare position with line. */
#define	ADC_AREF_DIFF	0.3333

#define	MOTOR_CHANGE	6000

/*
 * Init pins and assign values of necessary.
 */
void IO_init(void)
{
	DDRA &= ~(1 << PA0);	// Sensor 1
	DDRA &= ~(1 << PA1);	// Sensor 2
	DDRA &= ~(1 << PA2);	// Sensor 3
 
	DDRD |= (1 << PD0);		// Pin 1 Motor 1
	PORTD |= (1 << PD0);
	DDRD |= (1 << PD1);		// Pin 1 Motor 2
	PORTD |= (1 << PD1);
	DDRD |= (1 << PD4);		// Pin 2 Motor 1
	DDRD |= (1 << PD5);		// Pin 2 Motor 2

	DDRD |= (1 << PD7);
}

/*
 * Init Timer 1.
 * prescaler at 1
 * mode FPWM with top on ICR1
 * set on comp and clear on bottom
 */
void timer1_init()
{
	TCCR1B = (1 << WGM13) | (1 << WGM12) | (1 << CS10);
	TCCR1A = (1 << COM1A1) | (1 << COM1A0) | (1 << COM1B1) | (1 << COM1B0)
		| (1 << WGM11);

	ICR1 = 0x5000;

	OCR1A = ICR1;
	OCR1B = ICR1;
}

/*
 * Init Analog-Digital convertor.
 * parity at AVCC
 * no interrupt
 * prescaler at 32
 */
void ADC_init(void)
{
    ADMUX = (1 << REFS0);
    ADCSRA = (1 << ADEN) | (5 << ADPS0);
}

/*
 * Calculate tention (V) based on raw numerical value retrieved from ADC and
 * reference tention.
 */
double ADC_voltage(int raw)
{
    (void)raw;
    return raw * ADC_AREF_VOLTAGE / 1023.0;
}

/*
 * Retrieve raw numerica value from ADC.
 */
uint16_t ADC_get(uint8_t channel)
{
    ADMUX = (ADMUX & ~(0x1f << MUX0)) | channel;

    ADCSRA |= (1 << ADSC);
    while(ADCSRA & (1 << ADSC));

    return ADC;
}

/*
 * Check car position and decide speed of motors.
 */
void check()
{
	// Read values received from senzors to decide direction.
	double left = ADC_voltage(ADC_get(0));
	double right = ADC_voltage(ADC_get(2));

	if (left - right > ADC_AREF_VOLTAGE * 1.0 / 15) {
		// Car positioned to the right of the lign
		// Car should move left -> more speed for right motor.

		OCR1A = ICR1 - 2 * MOTOR_CHANGE;
		OCR1B = ICR1;
	} else if (right - left > ADC_AREF_VOLTAGE * 1.0 / 15) {
		// Car positioned to the left of the lign
		// Car should move right -> more speed for left motor.

		OCR1A = ICR1;
		OCR1B = ICR1 - 2 * MOTOR_CHANGE;
	} else {
		// Car position on the line -> motors move the same.

		OCR1A = ICR1;
		OCR1B = ICR1;
	}
}

/*
 * Application start point.
 */
int main (void)
{
	IO_init();
	timer1_init();
	ADC_init();

	sei();

	for(;;) {
		check();
		_delay_ms(5);
	}

	return 0;
}
